/*
 * Copyright (c) 2021-2022, Luca Fulchir <luker@fenrirproject.org>
 *
 * This file is part of dfim.
 *
 * dfim is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * dfim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with dfim.  If not, see <https://www.gnu.org/licenses/>.
 */

use super::DeviceMsg;
use super::WorkState;
use ::std::sync::Arc;
use ::tokio::sync::{mpsc, Mutex};

pub async fn drone(
    logger: Arc<::slog::Logger>,
    cfg: Arc<::libdfim::state::cfg::Cfg>,
    mtx_rx: Arc<Mutex<mpsc::UnboundedReceiver<DeviceMsg>>>,
) -> Result<(), ::std::io::Error> {
    ::slog::debug!(logger, "Running drone...");
    'unlocked: loop {
        // we use the same conditional variable to notify both new work and
        // a change in WORK_STATE

        let state =
            super::WORK_STATE.load(::std::sync::atomic::Ordering::Relaxed);
        match state {
            WorkState::KeepWorking
            | WorkState::ShutdownScannerGraceful
            | WorkState::ShutdownAllGraceful => {}
            WorkState::ShutdownImmediate => {
                break;
            }
        }
        let mut rx = mtx_rx.lock().await;
        let work = match rx.recv().await {
            Some(work) => {
                drop(rx);
                work
            }
            None => {
                ::slog::trace!(logger, "Drone exiting: no more senders");
                break;
            }
        };
        ::slog::debug!(logger, "Found device: {}- {:?}", work.id, work.path);
        let dev = ::udev::Device::from_syspath(&work.path).unwrap();
        let cfg_dev = cfg.devices.iter().find(|x| x.id == work.id).unwrap();
        use ::libdfim::config::trgt::TargetCommon;
        let cfg_target_id = cfg
            .targets
            .iter()
            .find(|x| x.id() == cfg_dev.target.id())
            .unwrap()
            .id();

        if cfg_dev.only_if_empty {
            match ::libdfim::state::is_empty(&dev) {
                Ok(is_empty) => {
                    if !is_empty {
                        ::slog::debug!(
                            logger,
                            "device not empty: \"{:?}\"",
                            work.path
                        );
                        continue 'unlocked;
                    }
                }
                Err(e) => {
                    ::slog::debug!(
                        logger,
                        "empty check: \"{:?}\" returned: {:?}",
                        work.path,
                        e
                    );
                    continue 'unlocked;
                }
            }
        }
        let check_state = match ::libdfim::state::get(
            &logger,
            &cfg,
            cfg_dev,
            cfg_target_id,
            &dev,
        ) {
            Ok(check_state) => check_state,
            Err(_) => {
                continue 'unlocked;
            }
        };
        match check_state {
            ::libdfim::state::s_trgt::CheckStatus::DoNotApply
            | ::libdfim::state::s_trgt::CheckStatus::AlreadyApplied => {
                // stop working on this device
                // either something needed to be empty and wasn't
                // or everything has already been applied
                ::slog::debug!(
                    logger,
                    "Dev {:?} returned {:?}",
                    dev.syspath(),
                    check_state
                );
                continue 'unlocked;
            }
            ::libdfim::state::s_trgt::CheckStatus::CanBeApplied => {}
        }
        match ::libdfim::state::set(&logger, &cfg, cfg_dev, cfg_target_id, &dev)
        {
            Ok(()) => {
                // all done and set
            }
            Err(_) => {
                // something went wrong on applying the conf.
                // nothing to be done unfortunately.
            }
        }
    }
    Ok(())
}
