/*
 * Copyright (c) 2021-2022, Luca Fulchir <luker@fenrirproject.org>
 *
 * This file is part of dfim.
 *
 * dfim is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * dfim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with dfim.  If not, see <https://www.gnu.org/licenses/>.
 */

mod drone;
mod scanner;

use ::std::sync::Arc;

pub struct DeviceMsg {
    id: String,
    path: ::std::path::PathBuf,
}
#[derive(Clone, Copy, Eq, PartialEq)]
pub enum WorkState {
    KeepWorking,
    ShutdownScannerGraceful,
    ShutdownAllGraceful,
    ShutdownImmediate,
}

/* Note:
 * Yes, we have a race condition on WORK_STATE
 * because we don't put it behind a mutex and it can be changed
 * by the signal handler and the scanner
 * It is as intended.
 * Pro:
 *   We don't lock mutexes every time we check WORK_STATE
 *   (which is a lot)
 * Con:
 *   you have to send a new ctrl-c signal because
 *   the shutdown is actually working and it overwrote your second ctrl-c
 */
pub static WORK_STATE: ::atomic::Atomic<WorkState> =
    ::atomic::Atomic::new(WorkState::KeepWorking);

pub async fn entry(
    logger: Arc<::slog::Logger>,
    arc_cfg: Arc<::libdfim::state::cfg::Cfg>,
) -> Result<(), ::std::io::Error> {
    let (tx, rx) = ::tokio::sync::mpsc::unbounded_channel::<DeviceMsg>();
    let arc_rx = Arc::new(::tokio::sync::Mutex::new(rx));

    let scanner_notify = Arc::new(::tokio::sync::Notify::new());

    let scan_sig = scanner_notify.clone();
    let sig_handler = ::tokio::spawn(async move {
        loop {
            match ::tokio::signal::ctrl_c().await {
                Err(_) => return,
                Ok(_) => {}
            }
            println!("got signal!");
            match WORK_STATE.load(::std::sync::atomic::Ordering::Relaxed) {
                WorkState::KeepWorking => WORK_STATE.store(
                    WorkState::ShutdownScannerGraceful,
                    ::std::sync::atomic::Ordering::Relaxed,
                ),
                _ => WORK_STATE.store(
                    WorkState::ShutdownImmediate,
                    ::std::sync::atomic::Ordering::Relaxed,
                ),
            };
            scan_sig.notify_waiters();
        }
    });

    let scanner_h = scanner::scanner(
        logger.clone(),
        scanner_notify.clone(),
        arc_cfg.clone(),
        tx,
    );
    let drone_h =
        ::tokio::spawn(drone::drone(logger.clone(), arc_cfg.clone(), arc_rx));

    let (scan_exit, drone_exit) = ::tokio::join!(scanner_h, drone_h);
    sig_handler.abort();
    match scan_exit {
        Ok(()) => match drone_exit {
            Ok(drone_res) => match drone_res {
                Ok(()) => Ok(()),
                Err(e) => Err(e),
            },
            Err(drone_join) => Err(drone_join.into()),
        },
        Err(scan_join) => Err(scan_join.into()),
    }
}
