/*
 * Copyright (c) 2021-2022, Luca Fulchir <luker@fenrirproject.org>
 *
 * This file is part of dfim.
 *
 * dfim is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * dfim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with dfim.  If not, see <https://www.gnu.org/licenses/>.
 */

use crate::config::errors;
use crate::config::trgt;

mod enums;
pub use enums::DevType;
pub use enums::Subsystem;

#[derive(::serde::Deserialize, ::serde::Serialize, Clone)]
#[serde(deny_unknown_fields)]
pub struct Attribute {
    pub name: String,
    #[serde(with = "serde_regex")]
    pub value: ::regex::Regex,
}
#[derive(::serde::Deserialize, ::serde::Serialize, Clone)]
#[serde(deny_unknown_fields)]
pub enum AttributeCheck {
    Ok(Attribute),
    Exclude(Attribute),
}
impl AttributeCheck {
    pub fn name(&self) -> &str {
        match self {
            AttributeCheck::Ok(a) => &a.name,
            AttributeCheck::Exclude(a) => &a.name,
        }
    }
}
#[derive(::serde::Deserialize, ::serde::Serialize, Clone)]
#[serde(deny_unknown_fields)]
pub struct Property {
    pub name: String,
    #[serde(with = "serde_regex")]
    pub value: ::regex::Regex,
}
#[derive(::serde::Deserialize, ::serde::Serialize, Clone)]
#[serde(deny_unknown_fields)]
pub enum PropertyCheck {
    Ok(Property),
    Exclude(Property),
}

impl PropertyCheck {
    pub fn name(&self) -> &str {
        match self {
            PropertyCheck::Ok(p) => &p.name,
            PropertyCheck::Exclude(p) => &p.name,
        }
    }
}

#[derive(::serde::Deserialize, ::serde::Serialize, Clone, Debug)]
#[serde(deny_unknown_fields)]
pub enum TagCheck {
    Ok(String),
    Exclude(String),
}
use std::ffi::OsStr;
impl AsRef<OsStr> for &TagCheck {
    fn as_ref(&self) -> &OsStr {
        match self {
            TagCheck::Ok(s) => OsStr::new(s),
            TagCheck::Exclude(s) => OsStr::new(s),
        }
    }
}

fn path_regex_default() -> ::regex::Regex {
    ::regex::Regex::new("^$").unwrap()
}

const fn only_if_empty_default() -> bool {
    true
}

#[derive(::serde::Deserialize, ::serde::Serialize, Clone)]
#[serde(deny_unknown_fields)]
pub struct Device {
    pub id: String,
    #[serde(default = "only_if_empty_default")]
    pub only_if_empty: bool,
    pub subsystem: Option<Subsystem>,
    pub devtype: Option<DevType>,
    #[serde(default = "path_regex_default")]
    #[serde(with = "serde_regex")]
    pub path_regex: ::regex::Regex,
    pub sysname: Option<String>,
    #[serde(default)]
    pub attributes: Vec<AttributeCheck>,
    #[serde(default)]
    pub properties: Vec<PropertyCheck>,
    #[serde(default)]
    pub tags: Vec<TagCheck>,
    #[serde(default)]
    pub current_tags: Vec<TagCheck>,
    pub target: trgt::TargetId,
}
impl super::CfgVerify for Device {
    fn standardize(&mut self) {
        // nothing to do
    }
    fn check_consistency(
        &self,
        _logger: &slog::Logger,
    ) -> Result<(), errors::ConfigError> {
        // do NOT check self.target.
        // That will be uninlined and checked later anyway
        // ...we don't have much to check at runtime here luckily
        Ok(())
    }
}
