/*
 * Copyright (c) 2021-2022, Luca Fulchir <luker@fenrirproject.org>
 *
 * This file is part of dfim.
 *
 * dfim is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * dfim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with dfim.  If not, see <https://www.gnu.org/licenses/>.
 */

macro_rules! partition_ids {
    (
        $(
            ($gptfdisk:expr, $uuid:expr, $desc:expr, $name:ident)$(,)*
        )+
    )
    => {
        #[derive(::serde::Deserialize, ::serde::Serialize, Clone, Copy, Debug,PartialEq)]
        #[serde(deny_unknown_fields)]
        pub enum PartType {
            $($name,)+
        }
        impl PartType {
            pub fn default_label(&self) -> String {
                match self {
                    $(PartType::$name => $desc.to_owned(),)+
                }
            }
        }
        impl From<PartType> for ::uuid::Uuid {
            fn from(p: PartType) -> ::uuid::Uuid {
                match p {
                    $(PartType::$name => ::uuid::Uuid::parse_str($uuid).unwrap(),)+
                }
            }
        }
        impl TryFrom<u16> for PartType {
            type Error = &'static str;
            fn try_from(value: u16) -> Result<Self, Self::Error> {
                match value {
                    $($gptfdisk => Ok(PartType::$name),)+
                    _ => Err("can't translate gptfdisk hex to uuid"),
                }
            }
        }
    };
}
/* These partition IDs/UUIDs are taken from
 * GPTFDISK source. report bugs there if you find them
 * here.
 * do not add new partitions if not supported from gptfdisk
 */
partition_ids!(
    (
        0x0000,
        "00000000-0000-0000-0000-000000000000",
        "Unused entry",
        Unused
    ),
    (
        0x0100,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft Basic Data",
        MicrosoftBasicData
    ),
    (
        0x0400,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        Fat16Less32M
    ),
    (
        0x0600,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        FAT16
    ),
    (
        0x0700,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        NTFSorHPFS
    ),
    (
        0x0701,
        "558D43C5-A1AC-43C0-AAC8-D1472B2923D1",
        "Microsoft Storage Replica",
        MicrosoftStorageReplica
    ),
    (
        0x0702,
        "90B6FF38-B98F-4358-A21F-48F35B4A8AD3",
        "ArcaOS Type 1",
        ArcaOSType1
    ),
    (
        0x0b00,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        Fat32
    ),
    (
        0x0c00,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        Fat32LBA
    ),
    (
        0x0c01,
        "E3C9E316-0B5C-4DB8-817D-F92DF00215AE",
        "Microsoft reserved",
        MicrosoftReserved
    ),
    (
        0x0e00,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        Fat16LBA
    ),
    (
        0x1100,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        HiddenFAT12
    ),
    (
        0x1400,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        HiddenFAT16less32M
    ),
    (
        0x1600,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        HiddenFAT16
    ),
    (
        0x1700,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        HiddenNTFSorHPFS
    ),
    (
        0x1b00,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        HiddenFAT32
    ),
    (
        0x1c00,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        HiddenFAT32LBA
    ),
    (
        0x1e00,
        "EBD0A0A2-B9E5-4433-87C0-68B6B72699C7",
        "Microsoft basic data",
        HiddenFAT16LBA
    ),
    (
        0x2700,
        "DE94BBA4-06D1-4D40-A16A-BFD50179D6AC",
        "Windows RE",
        WindowsRe
    ),
    (
        0x3000,
        "7412F7D5-A156-4B13-81DC-867174929325",
        "ONIE boot",
        OnieBoot
    ),
    (
        0x3001,
        "D4E6E2CD-4469-46F3-B5CB-1BFF57AFC149",
        "ONIE config",
        OnieConfig
    ),
    (
        0x3900,
        "C91818F9-8025-47AF-89D2-F030D7000C2C",
        "Plan 9",
        Plan9
    ),
    (
        0x4100,
        "9E1A2D38-C612-4316-AA26-8B49521E5A8B",
        "PowerPC PReP boot",
        PowerPcPRePBoot
    ),
    (
        0x4200,
        "AF9B60A0-1431-4F62-BC68-3311714A69AD",
        "Windows LDM data",
        WindowsLDMData
    ),
    (
        0x4201,
        "5808C8AA-7E8F-42E0-85D2-E1E90434CFB3",
        "Windows LDM metadata",
        WindowsLDMMetadata
    ),
    (
        0x4202,
        "E75CAF8F-F680-4CEE-AFA3-B001E56EFC2D",
        "Windows Storage Spaces",
        WindowsStorageSpaces
    ),
    (
        0x7501,
        "37AFFC90-EF7D-4E96-91C3-2D7AE055B174",
        "IBM GPFS",
        IbmGpfs
    ),
    (
        0x7f00,
        "FE3A2A5D-4F32-41A7-B725-ACCC3285A309",
        "ChromeOS kernel",
        ChromeOSKernel
    ),
    (
        0x7f01,
        "3CB8E202-3B7E-47DD-8A3C-7FF2A13CFCEC",
        "ChromeOS root",
        ChromeOSRoot
    ),
    (
        0x7f02,
        "2E0A753D-9E48-43B0-8337-B15192CB1B5E",
        "ChromeOS reserved",
        ChromeOSReserved
    ),
    (
        0x8200,
        "0657FD6D-A4AB-43C4-84E5-0933C84B4F4F",
        "Linux swap",
        LinuxSwap
    ),
    (
        0x8300,
        "0FC63DAF-8483-4772-8E79-3D69D8477DE4",
        "Linux filesystem",
        LinuxFilesystem
    ),
    (
        0x8301,
        "8DA63339-0007-60C0-C436-083AC8230908",
        "Linux reserved",
        LinuxReserved
    ),
    (
        0x8302,
        "933AC7E1-2EB4-4F13-B844-0E14E2AEF915",
        "Linux /home",
        LinuxHome
    ),
    (
        0x8303,
        "44479540-F297-41B2-9AF7-D131D5F0458A",
        "Linux x86 root (/)",
        Linuxx86Root
    ),
    (
        0x8304,
        "4F68BCE3-E8CD-4DB1-96E7-FBCAF984B709",
        "Linux x86-64 root (/)",
        Linuxx8664Root
    ),
    (
        0x8305,
        "B921B045-1DF0-41C3-AF44-4C6F280D3FAE",
        "Linux ARM64 root (/)",
        LinuxARM64Root
    ),
    (
        0x8306,
        "3B8F8425-20E0-4F3B-907F-1A25A76F98E8",
        "Linux /srv",
        LinuxSrv
    ),
    (
        0x8307,
        "69DAD710-2CE4-4E3C-B16C-21A1D49ABED3",
        "Linux ARM32 root (/)",
        LinuxARM32Root
    ),
    (
        0x8308,
        "7FFEC5C9-2D00-49B7-8941-3EA10A5586B7",
        "Linux dm-crypt",
        LinuxDmCrypt
    ),
    (
        0x8309,
        "CA7D7CCB-63ED-4C53-861C-1742536059CC",
        "Linux LUKS",
        LinuxLuks
    ),
    (
        0x830A,
        "993D8D3D-F80E-4225-855A-9DAF8ED7EA97",
        "Linux IA-64 root (/)",
        LinuxIA64Root
    ),
    (
        0x830B,
        "D13C5D3B-B5D1-422A-B29F-9454FDC89D76",
        "Linux x86 root verity",
        Linuxx86RootVerity
    ),
    (
        0x830C,
        "2C7357ED-EBD2-46D9-AEC1-23D437EC2BF5",
        "Linux x86-64 root verity",
        Linuxx8664RootVerity
    ),
    (
        0x830D,
        "7386CDF2-203C-47A9-A498-F2ECCE45A2D6",
        "Linux ARM32 root verity",
        LinuxARM32RootVerity
    ),
    (
        0x830E,
        "DF3300CE-D69F-4C92-978C-9BFB0F38D820",
        "Linux ARM64 root verity",
        LinuxARM64RootVerity
    ),
    (
        0x830F,
        "86ED10D5-B607-45BB-8957-D350F23D0571",
        "Linux IA-64 root verity",
        LinuxIA64RootVerity
    ),
    (
        0x8310,
        "4D21B016-B534-45C2-A9FB-5C16E091FD2D",
        "Linux /var",
        LinuxVar
    ),
    (
        0x8311,
        "7EC6F557-3BC5-4ACA-B293-16EF5DF639D1",
        "Linux /var/tmp",
        LinuxVarTmp
    ),
    (
        0x8312,
        "773F91EF-66D4-49B5-BD83-D683BF40AD16",
        "Linux user's home",
        LinuxUsersHome
    ),
    (
        0x8313,
        "75250D76-8CC6-458E-BD66-BD47CC81A812",
        "Linux x86 /usr",
        Linuxx86Usr
    ),
    (
        0x8314,
        "8484680C-9521-48C6-9C11-B0720656F69E",
        "Linux x86-64 /usr",
        Linuxx8664Usr
    ),
    (
        0x8315,
        "7D0359A3-02B3-4F0A-865C-654403E70625",
        "Linux ARM32 /usr",
        LinuxARM32Usr
    ),
    (
        0x8316,
        "B0E01050-EE5F-4390-949A-9101B17104E9",
        "Linux ARM64 /usr",
        LinuxARM64Usr
    ),
    (
        0x8317,
        "4301D2A6-4E3B-4B2A-BB94-9E0B2C4225EA",
        "Linux IA-64 /usr",
        LinuxIA64Usr
    ),
    (
        0x8318,
        "8F461B0D-14EE-4E81-9AA9-049B6FB97ABD",
        "Linux x86 /usr verity",
        Linuxx86UsrVerity
    ),
    (
        0x8319,
        "77FF5F63-E7B6-4633-ACF4-1565B864C0E6",
        "Linux x86-64 /usr verity",
        Linuxx8664UsrVerity
    ),
    (
        0x831A,
        "C215D751-7BCD-4649-BE90-6627490A4C05",
        "Linux ARM32 /usr verity",
        LinuxARM32UsrVerity
    ),
    (
        0x831B,
        "6E11A4E7-FBCA-4DED-B9E9-E1A512BB664E",
        "Linux ARM64 /usr verity",
        LinuxARM64UsrVerity
    ),
    (
        0x831C,
        "6A491E03-3BE7-4545-8E38-83320E0EA880",
        "Linux IA-64 /usr verity",
        LinuxIA64UsrVerity
    ),
    (
        0x8400,
        "D3BFE2DE-3DAF-11DF-BA40-E3A556D89593",
        "Intel Rapid Start",
        IntelRapidStart
    ),
    (
        0x8401,
        "7C5222BD-8F5D-4087-9C00-BF9843C7B58C",
        "SPDK block device",
        SpdkBlockDevice
    ),
    (
        0x8500,
        "5DFBF5F4-2848-4BAC-AA5E-0D9A20B745A6",
        "Container Linux /usr",
        ContainerLinuxUsr
    ),
    (
        0x8501,
        "3884DD41-8582-4404-B9A8-E9B84F2DF50E",
        "Container Linux resizable rootfs",
        ContainerLinuxResizableRootfs
    ),
    (
        0x8502,
        "C95DC21A-DF0E-4340-8D7B-26CBFA9A03E0",
        "Container Linux /OEM customizations",
        ContainerLinuxOEMCustomizations
    ),
    (
        0x8503,
        "BE9067B9-EA49-4F15-B4F6-F36F8C9E1818",
        "Container Linux root on RAID",
        ContainerLinuxRootOnRAID
    ),
    (
        0x8e00,
        "E6D6D379-F507-44C2-A23C-238F2A3DF928",
        "Linux LVM",
        LinuxLVM
    ),
    (
        0xa000,
        "2568845D-2332-4675-BC39-8FA5A4748D15",
        "Android bootloader",
        AndroidBootloader
    ),
    (
        0xa001,
        "114EAFFE-1552-4022-B26E-9B053604CF84",
        "Android bootloader 2",
        AndroidBootloader2
    ),
    (
        0xa002,
        "49A4D17F-93A3-45C1-A0DE-F50B2EBE2599",
        "Android boot 1",
        AndroidBoot1
    ),
    (
        0xa003,
        "4177C722-9E92-4AAB-8644-43502BFD5506",
        "Android recovery 1",
        AndroidRecovery1
    ),
    (
        0xa004,
        "EF32A33B-A409-486C-9141-9FFB711F6266",
        "Android misc",
        AndroidMisc
    ),
    (
        0xa005,
        "20AC26BE-20B7-11E3-84C5-6CFDB94711E9",
        "Android metadata",
        AndroidMetadata
    ),
    (
        0xa006,
        "38F428E6-D326-425D-9140-6E0EA133647C",
        "Android system 1",
        AndroidSystem1
    ),
    (
        0xa007,
        "A893EF21-E428-470A-9E55-0668FD91A2D9",
        "Android cache",
        AndroidCache1
    ),
    (
        0xa008,
        "DC76DDA9-5AC1-491C-AF42-A82591580C0D",
        "Android data",
        AndroidData
    ),
    (
        0xa009,
        "EBC597D0-2053-4B15-8B64-E0AAC75F4DB1",
        "Android persistent",
        AndroidPersistent
    ),
    (
        0xa00a,
        "8F68CC74-C5E5-48DA-BE91-A0C8C15E9C80",
        "Android factory",
        AndroidFactory
    ),
    (
        0xa00b,
        "767941D0-2085-11E3-AD3B-6CFDB94711E9",
        "Android fastboot/tertiary",
        AndroidFastbootTertiary
    ),
    (
        0xa00c,
        "AC6D7924-EB71-4DF8-B48D-E267B27148FF",
        "Android OEM",
        AndroidOEM
    ),
    (
        0xa00d,
        "C5A0AEEC-13EA-11E5-A1B1-001E67CA0C3C",
        "Android vendor",
        AndroidVendor
    ),
    (
        0xa00e,
        "BD59408B-4514-490D-BF12-9878D963F378",
        "Android config",
        AndroidConfig
    ),
    (
        0xa00f,
        "9FDAA6EF-4B3F-40D2-BA8D-BFF16BFB887B",
        "Android factory (alt)",
        AndroidFactoryAlt
    ),
    (
        0xa010,
        "19A710A2-B3CA-11E4-B026-10604B889DCF",
        "Android meta",
        AndroidMeta
    ),
    (
        0xa011,
        "193D1EA4-B3CA-11E4-B075-10604B889DCF",
        "Android EXT",
        AndroidExt
    ),
    (
        0xa012,
        "DEA0BA2C-CBDD-4805-B4F9-F428251C3E98",
        "Android SBL1",
        AndroidSBL1
    ),
    (
        0xa013,
        "8C6B52AD-8A9E-4398-AD09-AE916E53AE2D",
        "Android SBL2",
        AndroidSBL2
    ),
    (
        0xa014,
        "05E044DF-92F1-4325-B69E-374A82E97D6E",
        "Android SBL3",
        AndroidSBL3
    ),
    (
        0xa015,
        "400FFDCD-22E0-47E7-9A23-F16ED9382388",
        "Android APPSBL",
        AndroidAPPSBL
    ),
    (
        0xa016,
        "A053AA7F-40B8-4B1C-BA08-2F68AC71A4F4",
        "Android QSEE/tz",
        AndroidQSEETz
    ),
    (
        0xa017,
        "E1A6A689-0C8D-4CC6-B4E8-55A4320FBD8A",
        "Android QHEE/hyp",
        AndroidQHEEhyp
    ),
    (
        0xa018,
        "098DF793-D712-413D-9D4E-89D711772228",
        "Android RPM",
        AndroidRPM
    ),
    (
        0xa019,
        "D4E0D938-B7FA-48C1-9D21-BC5ED5C4B203",
        "Android WDOG debug/sdi",
        AndroidWDOGDebugSdi
    ),
    (
        0xa01a,
        "20A0C19C-286A-42FA-9CE7-F64C3226A794",
        "Android DDR",
        AndroidDDR
    ),
    (
        0xa01b,
        "A19F205F-CCD8-4B6D-8F1E-2D9BC24CFFB1",
        "Android CDT",
        AndroidCDT
    ),
    (
        0xa01c,
        "66C9B323-F7FC-48B6-BF96-6F32E335A428",
        "Android RAM dump",
        AndroidRAMDump
    ),
    (
        0xa01d,
        "303E6AC3-AF15-4C54-9E9B-D9A8FBECF401",
        "Android SEC",
        AndroidSEC
    ),
    (
        0xa01e,
        "C00EEF24-7709-43D6-9799-DD2B411E7A3C",
        "Android PMIC",
        AndroidPMIC
    ),
    (
        0xa01f,
        "82ACC91F-357C-4A68-9C8F-689E1B1A23A1",
        "Android misc 1",
        AndroidMisc1
    ),
    (
        0xa020,
        "E2802D54-0545-E8A1-A1E8-C7A3E245ACD4",
        "Android misc 2",
        AndroidMisc2
    ),
    (
        0xa021,
        "65ADDCF4-0C5C-4D9A-AC2D-D90B5CBFCD03",
        "Android device info",
        AndroidDeviceInfo
    ),
    (
        0xa022,
        "E6E98DA2-E22A-4D12-AB33-169E7DEAA507",
        "Android APDP",
        AndroidAPDP
    ),
    (
        0xa023,
        "ED9E8101-05FA-46B7-82AA-8D58770D200B",
        "Android MSADP",
        AndroidMSADP
    ),
    (
        0xa024,
        "11406F35-1173-4869-807B-27DF71802812",
        "Android DPO",
        AndroidDPO
    ),
    (
        0xa025,
        "9D72D4E4-9958-42DA-AC26-BEA7A90B0434",
        "Android recovery 2",
        AndroidRecovery2
    ),
    (
        0xa026,
        "6C95E238-E343-4BA8-B489-8681ED22AD0B",
        "Android persist",
        AndroidPersist
    ),
    (
        0xa027,
        "EBBEADAF-22C9-E33B-8F5D-0E81686A68CB",
        "Android modem ST1",
        AndroidModemST1
    ),
    (
        0xa028,
        "0A288B1F-22C9-E33B-8F5D-0E81686A68CB",
        "Android modem ST2",
        AndroidModemST2
    ),
    (
        0xa029,
        "57B90A16-22C9-E33B-8F5D-0E81686A68CB",
        "Android FSC",
        AndroidFSC
    ),
    (
        0xa02a,
        "638FF8E2-22C9-E33B-8F5D-0E81686A68CB",
        "Android FSG 1",
        AndroidFSG1
    ),
    (
        0xa02b,
        "2013373E-1AC4-4131-BFD8-B6A7AC638772",
        "Android FSG 2",
        AndroidFSG2
    ),
    (
        0xa02c,
        "2C86E742-745E-4FDD-BFD8-B6A7AC638772",
        "Android SSD",
        AndroidSSD
    ),
    (
        0xa02d,
        "DE7D4029-0F5B-41C8-AE7E-F6C023A02B33",
        "Android keystore",
        AndroidKeystore
    ),
    (
        0xa02e,
        "323EF595-AF7A-4AFA-8060-97BE72841BB9",
        "Android encrypt",
        AndroidEncrypt
    ),
    (
        0xa02f,
        "45864011-CF89-46E6-A445-85262E065604",
        "Android EKSST",
        AndroidEKSST
    ),
    (
        0xa030,
        "8ED8AE95-597F-4C8A-A5BD-A7FF8E4DFAA9",
        "Android RCT",
        AndroidRCT
    ),
    (
        0xa031,
        "DF24E5ED-8C96-4B86-B00B-79667DC6DE11",
        "Android spare1",
        AndroidSpare1
    ),
    (
        0xa032,
        "7C29D3AD-78B9-452E-9DEB-D098D542F092",
        "Android spare2",
        AndroidSpare2
    ),
    (
        0xa033,
        "379D107E-229E-499D-AD4F-61F5BCF87BD4",
        "Android spare3",
        AndroidSpare3
    ),
    (
        0xa034,
        "0DEA65E5-A676-4CDF-823C-77568B577ED5",
        "Android spare4",
        AndroidSpare4
    ),
    (
        0xa035,
        "4627AE27-CFEF-48A1-88FE-99C3509ADE26",
        "Android raw resources",
        AndroidRawResources
    ),
    (
        0xa036,
        "20117F86-E985-4357-B9EE-374BC1D8487D",
        "Android boot 2",
        AndroidBoot2
    ),
    (
        0xa037,
        "86A7CB80-84E1-408C-99AB-694F1A410FC7",
        "Android FOTA",
        AndroidFOTA
    ),
    (
        0xa038,
        "97D7B011-54DA-4835-B3C4-917AD6E73D74",
        "Android system 2",
        AndroidSystem2
    ),
    (
        0xa039,
        "5594C694-C871-4B5F-90B1-690A6F68E0F7",
        "Android cache",
        AndroidCache2
    ),
    (
        0xa03a,
        "1B81E7E6-F50D-419B-A739-2AEEF8DA3335",
        "Android user data",
        AndroidUserData
    ),
    (
        0xa03b,
        "98523EC6-90FE-4C67-B50A-0FC59ED6F56D",
        "LG (Android) advanced flasher",
        LGAndroidAdvancedFlasher
    ),
    (
        0xa03c,
        "2644BCC0-F36A-4792-9533-1738BED53EE3",
        "Android PG1FS",
        AndroidPG1FS
    ),
    (
        0xa03d,
        "DD7C91E9-38C9-45C5-8A12-4A80F7E14057",
        "Android PG2FS",
        AndroidPG2FS
    ),
    (
        0xa03e,
        "7696D5B6-43FD-4664-A228-C563C4A1E8CC",
        "Android board info",
        AndroidBoardInfo
    ),
    (
        0xa03f,
        "0D802D54-058D-4A20-AD2D-C7A362CEACD4",
        "Android MFG",
        AndroidMFG
    ),
    (
        0xa040,
        "10A0C19C-516A-5444-5CE3-664C3226A794",
        "Android limits",
        AndroidLimits
    ),
    (
        0xa200,
        "734E5AFE-F61A-11E6-BC64-92361F002671",
        "Atari TOS basic data",
        AtariTOSBasicData
    ),
    (
        0xa500,
        "516E7CB4-6ECF-11D6-8FF8-00022D09712B",
        "FreeBSD disklabel",
        FreeBSDDisklabel
    ),
    (
        0xa501,
        "83BD6B9D-7F41-11DC-BE0B-001560B84F0F",
        "FreeBSD boot",
        FreeBSDBoot
    ),
    (
        0xa502,
        "516E7CB5-6ECF-11D6-8FF8-00022D09712B",
        "FreeBSD swap",
        FreeBSDSwap
    ),
    (
        0xa503,
        "516E7CB6-6ECF-11D6-8FF8-00022D09712B",
        "FreeBSD UFS",
        FreeBSDUFS
    ),
    (
        0xa504,
        "516E7CBA-6ECF-11D6-8FF8-00022D09712B",
        "FreeBSD ZFS",
        FreeBSDZFS
    ),
    (
        0xa505,
        "516E7CB8-6ECF-11D6-8FF8-00022D09712B",
        "FreeBSD Vinum/RAID",
        FreeBSDVinumRAID
    ),
    (
        0xa580,
        "85D5E45A-237C-11E1-B4B3-E89A8F7FC3A7",
        "Midnight BSD data",
        MidnightBSDData
    ),
    (
        0xa581,
        "85D5E45E-237C-11E1-B4B3-E89A8F7FC3A7",
        "Midnight BSD boot",
        MidnightBSDBoot
    ),
    (
        0xa582,
        "85D5E45B-237C-11E1-B4B3-E89A8F7FC3A7",
        "Midnight BSD swap",
        MidnightBSDSwap
    ),
    (
        0xa583,
        "0394Ef8B-237E-11E1-B4B3-E89A8F7FC3A7",
        "Midnight BSD UFS",
        MidnightBSDUFS
    ),
    (
        0xa584,
        "85D5E45D-237C-11E1-B4B3-E89A8F7FC3A7",
        "Midnight BSD ZFS",
        MidnightBSDZFS
    ),
    (
        0xa585,
        "85D5E45C-237C-11E1-B4B3-E89A8F7FC3A7",
        "Midnight BSD Vinum",
        MidnightBSDVinum
    ),
    (
        0xa600,
        "824CC7A0-36A8-11E3-890A-952519AD3F61",
        "OpenBSD disklabel",
        OpenBSDDisklabel
    ),
    (
        0xa800,
        "55465300-0000-11AA-AA11-00306543ECAC",
        "Apple UFS",
        AppleUFS
    ),
    (
        0xa900,
        "516E7CB4-6ECF-11D6-8FF8-00022D09712B",
        "FreeBSD disklabel",
        FreeBSDNETBSDDisklabel
    ),
    (
        0xa901,
        "49F48D32-B10E-11DC-B99B-0019D1879648",
        "NetBSD swap",
        NetBSDSwap
    ),
    (
        0xa902,
        "49F48D5A-B10E-11DC-B99B-0019D1879648",
        "NetBSD FFS",
        NetBSDFFS
    ),
    (
        0xa903,
        "49F48D82-B10E-11DC-B99B-0019D1879648",
        "NetBSD LFS",
        NetBSDLFS
    ),
    (
        0xa904,
        "2DB519C4-B10F-11DC-B99B-0019D1879648",
        "NetBSD concatenated",
        NetBSDConcatenated
    ),
    (
        0xa905,
        "2DB519EC-B10F-11DC-B99B-0019D1879648",
        "NetBSD encrypted",
        NetBSDEncrypted
    ),
    (
        0xa906,
        "49F48DAA-B10E-11DC-B99B-0019D1879648",
        "NetBSD RAID",
        NetBSDRAID
    ),
    (
        0xab00,
        "426F6F74-0000-11AA-AA11-00306543ECAC",
        "Recovery HD",
        RecoveryHD
    ),
    (
        0xaf00,
        "48465300-0000-11AA-AA11-00306543ECAC",
        "Apple HFS/HFS+",
        AppleHfsHfsPlus
    ),
    (
        0xaf01,
        "52414944-0000-11AA-AA11-00306543ECAC",
        "Apple RAID",
        AppleRAID
    ),
    (
        0xaf02,
        "52414944-5F4F-11AA-AA11-00306543ECAC",
        "Apple RAID offline",
        AppleRAIDOffline
    ),
    (
        0xaf03,
        "4C616265-6C00-11AA-AA11-00306543ECAC",
        "Apple label",
        AppleLabel
    ),
    (
        0xaf04,
        "5265636F-7665-11AA-AA11-00306543ECAC",
        "AppleTV recovery",
        AppleTVRecovery
    ),
    (
        0xaf05,
        "53746F72-6167-11AA-AA11-00306543ECAC",
        "Apple Core Storage",
        AppleCoreStorage
    ),
    (
        0xaf06,
        "B6FA30DA-92D2-4A9A-96F1-871EC6486200",
        "Apple SoftRAID Status",
        AppleSoftRAIDStatus
    ),
    (
        0xaf07,
        "2E313465-19B9-463F-8126-8A7993773801",
        "Apple SoftRAID Scratch",
        AppleSoftRAIDScratch
    ),
    (
        0xaf08,
        "FA709C7E-65B1-4593-BFD5-E71D61DE9B02",
        "Apple SoftRAID Volume",
        AppleSoftRAIDVolume
    ),
    (
        0xaf09,
        "BBBA6DF5-F46F-4A89-8F59-8765B2727503",
        "Apple SoftRAID Cache",
        AppleSoftRAIDCache
    ),
    (
        0xaf0a,
        "7C3457EF-0000-11AA-AA11-00306543ECAC",
        "Apple APFS",
        AppleAPFS
    ),
    (
        0xb300,
        "CEF5A9AD-73BC-4601-89F3-CDEEEEE321A1",
        "QNX6 Power-Safe",
        QNX6PowerSafe
    ),
    (
        0xbb00,
        "4778ED65-BF42-45FA-9C5B-287A1DC4AAB1",
        "Barebox boot loader",
        BareboxBootLoader
    ),
    (
        0xbc00,
        "0311FC50-01CA-4725-AD77-9ADBB20ACE98",
        "Acronis Secure Zone",
        AcronisSecureZone
    ),
    (
        0xbe00,
        "6A82CB45-1DD2-11B2-99A6-080020736631",
        "Solaris boot",
        SolarisBoot
    ),
    (
        0xbf00,
        "6A85CF4D-1DD2-11B2-99A6-080020736631",
        "Solaris root",
        SolarisRoot
    ),
    (
        0xbf01,
        "6A898CC3-1DD2-11B2-99A6-080020736631",
        "Solaris /usr & Mac ZFS",
        SolarisUsrMacZFS
    ),
    (
        0xbf02,
        "6A87C46F-1DD2-11B2-99A6-080020736631",
        "Solaris swap",
        SolarisSwap
    ),
    (
        0xbf03,
        "6A8B642B-1DD2-11B2-99A6-080020736631",
        "Solaris backup",
        SolarisBackup
    ),
    (
        0xbf04,
        "6A8EF2E9-1DD2-11B2-99A6-080020736631",
        "Solaris /var",
        SolarisVar
    ),
    (
        0xbf05,
        "6A90BA39-1DD2-11B2-99A6-080020736631",
        "Solaris /home",
        SolarisHome
    ),
    (
        0xbf06,
        "6A9283A5-1DD2-11B2-99A6-080020736631",
        "Solaris alternate sector",
        SolarisAlternateSector
    ),
    (
        0xbf07,
        "6A945A3B-1DD2-11B2-99A6-080020736631",
        "Solaris Reserved 1",
        SolarisReserved1
    ),
    (
        0xbf08,
        "6A9630D1-1DD2-11B2-99A6-080020736631",
        "Solaris Reserved 2",
        SolarisReserved2
    ),
    (
        0xbf09,
        "6A980767-1DD2-11B2-99A6-080020736631",
        "Solaris Reserved 3",
        SolarisReserved3
    ),
    (
        0xbf0a,
        "6A96237F-1DD2-11B2-99A6-080020736631",
        "Solaris Reserved 4",
        SolarisReserved4
    ),
    (
        0xbf0b,
        "6A8D2AC7-1DD2-11B2-99A6-080020736631",
        "Solaris Reserved 5",
        SolarisReserved5
    ),
    (
        0xc001,
        "75894C1E-3AEB-11D3-B7C1-7B03A0000000",
        "HP-UX data",
        HPUXData
    ),
    (
        0xc002,
        "E2A1E728-32E3-11D6-A682-7B03A0000000",
        "HP-UX service",
        HPUXService
    ),
    (
        0xe100,
        "7412F7D5-A156-4B13-81DC-867174929325",
        "ONIE boot",
        OnieBoot2
    ),
    (
        0xe101,
        "D4E6E2CD-4469-46F3-B5CB-1BFF57AFC149",
        "ONIE config",
        OnieConfig2
    ),
    (
        0xe900,
        "8C8F8EFF-AC95-4770-814A-21994F2DBC8F",
        "Veracrypt data",
        VeracryptData
    ),
    (
        0xea00,
        "BC13C2FF-59E6-4262-A352-B275FD6F7172",
        "XBOOTLDR partition",
        XBOOTLDRPartition
    ),
    (
        0xeb00,
        "42465331-3BA3-10F1-802A-4861696B7521",
        "Haiku BFS",
        HaikuBFS
    ),
    (
        0xed00,
        "F4019732-066E-4E12-8273-346C5641494F",
        "Sony system partition",
        SonySystemPartition
    ),
    (
        0xed01,
        "BFBFAFE7-A34F-448A-9A5B-6213EB736C22",
        "Lenovo system partition",
        LenovoSystemPartition
    ),
    (
        0xef00,
        "C12A7328-F81F-11D2-BA4B-00A0C93EC93B",
        "EFI system partition",
        EFISystemPartition
    ),
    (
        0xef01,
        "024DEE41-33E7-11D3-9D69-0008C781F39F",
        "MBR partition scheme",
        MBRPartitionScheme
    ),
    (
        0xef02,
        "21686148-6449-6E6F-744E-656564454649",
        "BIOS boot partition",
        BIOSBootPartition
    ),
    (
        0xf800,
        "4FBD7E29-9D25-41B8-AFD0-062C0CEFF05D",
        "Ceph OSD",
        CephOSD
    ),
    (
        0xf801,
        "4FBD7E29-9D25-41B8-AFD0-5EC00CEFF05D",
        "Ceph dm-crypt OSD",
        CephDmCryptOSD
    ),
    (
        0xf802,
        "45B0969E-9B03-4F30-B4C6-B4B80CEFF106",
        "Ceph journal",
        CephJournal
    ),
    (
        0xf803,
        "45B0969E-9B03-4F30-B4C6-5EC00CEFF106",
        "Ceph dm-crypt journal",
        CephDmCryptJournal
    ),
    (
        0xf804,
        "89C57F98-2FE5-4DC0-89C1-F3AD0CEFF2BE",
        "Ceph disk in creation",
        CephDiskInCreation
    ),
    (
        0xf805,
        "89C57F98-2FE5-4DC0-89C1-5EC00CEFF2BE",
        "Ceph dm-crypt disk in creation",
        CephDmCryptDiskInCreation
    ),
    (
        0xf806,
        "CAFECAFE-9B03-4F30-B4C6-B4B80CEFF106",
        "Ceph block",
        CephBlock
    ),
    (
        0xf807,
        "30CD0809-C2B2-499C-8879-2D6B78529876",
        "Ceph block DB",
        CephBlockDB
    ),
    (
        0xf808,
        "5CE17FCE-4087-4169-B7FF-056CC58473F9",
        "Ceph block write-ahead log",
        CephBlockWriteAheadLog
    ),
    (
        0xf809,
        "FB3AABF9-D25F-47CC-BF5E-721D1816496B",
        "Ceph lockbox for dm-crypt keys",
        CephLockboxForDmCryptKeys
    ),
    (
        0xf80a,
        "4FBD7E29-8AE0-4982-BF9D-5A8D867AF560",
        "Ceph multipath OSD",
        CephMultipathOSD
    ),
    (
        0xf80b,
        "45B0969E-8AE0-4982-BF9D-5A8D867AF560",
        "Ceph multipath journal",
        CephMultipathJournal
    ),
    (
        0xf80c,
        "CAFECAFE-8AE0-4982-BF9D-5A8D867AF560",
        "Ceph multipath block 1",
        CephMultipathBlock1
    ),
    (
        0xf80d,
        "7F4A666A-16F3-47A2-8445-152EF4D03F6C",
        "Ceph multipath block 2",
        CephMultipathBlock2
    ),
    (
        0xf80e,
        "EC6D6385-E346-45DC-BE91-DA2A7C8B3261",
        "Ceph multipath block DB",
        CephMultipathBlockDB
    ),
    (
        0xf80f,
        "01B41E1B-002A-453C-9F17-88793989FF8F",
        "Ceph multipath block write-ahead log",
        CephMultipathBlockWriteAheadLog
    ),
    (
        0xf810,
        "CAFECAFE-9B03-4F30-B4C6-5EC00CEFF106",
        "Ceph dm-crypt block",
        CephDmCryptBlock
    ),
    (
        0xf811,
        "93B0052D-02D9-4D8A-A43B-33A3EE4DFBC3",
        "Ceph dm-crypt block DB",
        CephDmCryptBlockDB
    ),
    (
        0xf812,
        "306E8683-4FE2-4330-B7C0-00A917C16966",
        "Ceph dm-crypt block write-ahead log",
        CephDmCryptBlockWriteAheadLog
    ),
    (
        0xf813,
        "45B0969E-9B03-4F30-B4C6-35865CEFF106",
        "Ceph dm-crypt LUKS journal",
        CephDmCryptLUKSJournal
    ),
    (
        0xf814,
        "CAFECAFE-9B03-4F30-B4C6-35865CEFF106",
        "Ceph dm-crypt LUKS block",
        CephDmCryptLUKSblock
    ),
    (
        0xf815,
        "166418DA-C469-4022-ADF4-B30AFD37F176",
        "Ceph dm-crypt LUKS block DB",
        CephDmCryptLUKSBlockDB
    ),
    (
        0xf816,
        "86A32090-3647-40B9-BBBD-38D8C573AA86",
        "Ceph dm-crypt LUKS block write-ahead log",
        CephDmCryptLUKSBlockWriteAheadLog
    ),
    (
        0xf817,
        "4FBD7E29-9D25-41B8-AFD0-35865CEFF05D",
        "Ceph dm-crypt LUKS OSD",
        CephDmCryptLUKSOSD
    ),
    (
        0xfb00,
        "AA31E02A-400F-11DB-9590-000C2911D1B8",
        "VMWare VMFS",
        VMWareVMFS
    ),
    (
        0xfb01,
        "9198EFFC-31C0-11DB-8F78-000C2911D1B8",
        "VMWare reserved",
        VMWareReserved
    ),
    (
        0xfc00,
        "9D275380-40AD-11DB-BF97-000C2911D1B8",
        "VMWare kcore crash protection",
        VMWareKcoreCrashProtection
    ),
    (
        0xfd00,
        "A19D880F-05FC-4D3B-A006-743F0F84911E",
        "Linux RAID",
        LinuxRAID
    ),
);
