/*
 * Copyright (c) 2021-2022, Luca Fulchir <luker@fenrirproject.org>
 *
 * This file is part of dfim.
 *
 * dfim is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * dfim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with dfim.  If not, see <https://www.gnu.org/licenses/>.
 */

use slog::Drain;

pub type Error = ::std::io::Error;

pub fn drop_logger(_logger: ::std::sync::Arc<::slog::Logger>) {
    // nothing, just force dropping
    // we need to force-drop the loggger since it's async
    // and the ::std::process::exit(..) call does not call any destructor
}

pub fn create_default_logger() -> ::slog::Logger {
    let decorator = ::slog_term::TermDecorator::new();
    let decorator = decorator.stderr();
    let drain_term = ::slog_term::FullFormat::new(decorator.build());
    let drain_term = drain_term.use_utc_timestamp();
    let drain_async = ::slog_async::Async::new(drain_term.build().fuse())
        .build()
        .filter_level(::slog::Level::Info)
        .fuse();

    ::slog::Logger::root(drain_async, ::slog::o!())
}

pub fn create_logger(
    conf_logger: &crate::config::Logging,
) -> Result<::slog::Logger, Error> {
    match conf_logger.dest {
        crate::config::LogDest::Terminal(_) => setup_terminal(conf_logger),
        crate::config::LogDest::Syslog { .. } => setup_syslog(conf_logger),
    }
}

fn setup_terminal(
    conf_logger: &crate::config::Logging,
) -> Result<::slog::Logger, Error> {
    // set stdout or stderr
    let decorator = ::slog_term::TermDecorator::new();
    let decorator = match conf_logger.dest {
        crate::config::LogDest::Terminal(opt_stream) => match opt_stream {
            None => decorator,
            Some(stream) => match stream {
                crate::config::LogTermStream::StdErr => decorator.stderr(),
                crate::config::LogTermStream::StdOut => decorator.stdout(),
            },
        },
        _ => ::std::panic!("logger setup: expected terminal, got syslog"),
    };
    let drain_term = ::slog_term::FullFormat::new(decorator.build());
    let drain_term = match conf_logger.time {
        crate::config::LogTime::UTC => drain_term.use_utc_timestamp(),
        crate::config::LogTime::Local => drain_term.use_local_timestamp(),
    };
    let drain_async = ::slog_async::Async::new(drain_term.build().fuse())
        .build()
        .filter_level(conf_logger.level.into())
        .fuse();

    Ok(::slog::Logger::root(drain_async, ::slog::o!()))
}

fn setup_syslog(
    conf_logger: &crate::config::Logging,
) -> Result<::slog::Logger, Error> {
    let logger = ::slog_syslog::SyslogBuilder::new();
    let (syslog_dest, syslog_facility, syslog_level) = match &conf_logger.dest {
        crate::config::LogDest::Syslog {
            dest,
            facility,
            level,
        } => (dest, *facility, *level),
        _ => ::std::panic!("logger setup: expected terminal, got syslog"),
    };

    let logger = match syslog_dest {
        crate::config::LogSyslog::Socket(path) => logger.unix(path),
        crate::config::LogSyslog::Udp {
            from,
            to,
            hostname_id,
        } => logger.udp(*from, *to, hostname_id),
        crate::config::LogSyslog::Tcp { to, hostname_id } => {
            logger.tcp(*to, hostname_id)
        }
    };

    let logger = logger
        .facility(syslog_facility.into())
        .level(syslog_level.into())
        .start()?;

    let logger = logger.filter_level(conf_logger.level.into());

    Ok(::slog::Logger::root(logger.fuse(), ::slog::o!()))
}

impl From<crate::config::LogLevel> for ::slog::Level {
    fn from(level: crate::config::LogLevel) -> Self {
        match level {
            crate::config::LogLevel::Critical => ::slog::Level::Critical,
            crate::config::LogLevel::Error => ::slog::Level::Error,
            crate::config::LogLevel::Warning => ::slog::Level::Warning,
            crate::config::LogLevel::Info => ::slog::Level::Info,
            crate::config::LogLevel::Debug => ::slog::Level::Debug,
            crate::config::LogLevel::Trace => ::slog::Level::Trace,
        }
    }
}

impl From<crate::config::LogFacility> for ::slog_syslog::Facility {
    fn from(level: crate::config::LogFacility) -> Self {
        match level {
            crate::config::LogFacility::Kern => {
                ::slog_syslog::Facility::LOG_KERN
            }
            crate::config::LogFacility::User => {
                ::slog_syslog::Facility::LOG_USER
            }
            crate::config::LogFacility::Mail => {
                ::slog_syslog::Facility::LOG_MAIL
            }
            crate::config::LogFacility::Daemon => {
                ::slog_syslog::Facility::LOG_DAEMON
            }
            crate::config::LogFacility::Auth => {
                ::slog_syslog::Facility::LOG_AUTH
            }
            crate::config::LogFacility::Syslog => {
                ::slog_syslog::Facility::LOG_SYSLOG
            }
            crate::config::LogFacility::Lpr => ::slog_syslog::Facility::LOG_LPR,
            crate::config::LogFacility::News => {
                ::slog_syslog::Facility::LOG_NEWS
            }
            crate::config::LogFacility::Uucp => {
                ::slog_syslog::Facility::LOG_UUCP
            }
            crate::config::LogFacility::Cron => {
                ::slog_syslog::Facility::LOG_CRON
            }
            crate::config::LogFacility::Authpriv => {
                ::slog_syslog::Facility::LOG_AUTHPRIV
            }
            crate::config::LogFacility::Ftp => ::slog_syslog::Facility::LOG_FTP,
            crate::config::LogFacility::Local0 => {
                ::slog_syslog::Facility::LOG_LOCAL0
            }
            crate::config::LogFacility::Local1 => {
                ::slog_syslog::Facility::LOG_LOCAL1
            }
            crate::config::LogFacility::Local2 => {
                ::slog_syslog::Facility::LOG_LOCAL2
            }
            crate::config::LogFacility::Local3 => {
                ::slog_syslog::Facility::LOG_LOCAL3
            }
            crate::config::LogFacility::Local4 => {
                ::slog_syslog::Facility::LOG_LOCAL4
            }
            crate::config::LogFacility::Local5 => {
                ::slog_syslog::Facility::LOG_LOCAL5
            }
            crate::config::LogFacility::Local6 => {
                ::slog_syslog::Facility::LOG_LOCAL6
            }
            crate::config::LogFacility::Local7 => {
                ::slog_syslog::Facility::LOG_LOCAL7
            }
        }
    }
}
