/*
 * Copyright (c) 2021-2022, Luca Fulchir <luker@fenrirproject.org>
 *
 * This file is part of dfim.
 *
 * dfim is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * dfim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with dfim.  If not, see <https://www.gnu.org/licenses/>.
 */

pub mod s_exec;
pub mod s_fs;
pub mod s_gpt;

use crate::config;

pub enum Target {
    Dummy,
    Exec(s_exec::Exec),
    GPT(s_gpt::GPT),
    BTRFS(s_fs::btrfs::BTRFS),
}

#[derive(::serde::Deserialize, Clone, Debug)]
#[serde(deny_unknown_fields)]
pub enum CheckStatus {
    DoNotApply,
    CanBeApplied,
    AlreadyApplied,
}

pub struct NextTarget {
    pub id: String,
    pub dev: ::udev::Device,
}

pub struct GetCurrentResult {
    pub status: CheckStatus,
    pub conf: crate::config::trgt::Target,
    pub next: Vec<NextTarget>,
}

pub trait TargetApply {
    fn dev_to_config(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<crate::config::trgt::Target, ::std::io::Error>;
    fn get_current(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<GetCurrentResult, ::std::io::Error>;
    fn apply(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<Vec<NextTarget>, ::std::io::Error>;
}

impl TargetApply for Target {
    fn dev_to_config(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<crate::config::trgt::Target, ::std::io::Error> {
        match self {
            Target::Dummy => Ok(crate::config::trgt::Target::Dummy),
            Target::Exec(e) => e.dev_to_config(logger, cfg, cfg_dev, dev),
            Target::GPT(g) => g.dev_to_config(logger, cfg, cfg_dev, dev),
            Target::BTRFS(b) => b.dev_to_config(logger, cfg, cfg_dev, dev),
        }
    }
    fn get_current(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<GetCurrentResult, ::std::io::Error> {
        match self {
            Target::Dummy => Ok(GetCurrentResult {
                status: CheckStatus::AlreadyApplied,
                conf: crate::config::trgt::Target::Dummy,
                next: Vec::new(),
            }),
            Target::Exec(e) => e.get_current(logger, cfg, cfg_dev, dev),
            Target::GPT(g) => g.get_current(logger, cfg, cfg_dev, dev),
            Target::BTRFS(b) => b.get_current(logger, cfg, cfg_dev, dev),
        }
    }
    fn apply(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<Vec<NextTarget>, ::std::io::Error> {
        match self {
            Target::Dummy => Ok(Vec::new()),
            Target::Exec(e) => e.apply(logger, cfg, cfg_dev, dev),
            Target::GPT(g) => g.apply(logger, cfg, cfg_dev, dev),
            Target::BTRFS(b) => b.apply(logger, cfg, cfg_dev, dev),
        }
    }
}

impl From<&crate::config::trgt::Target> for Target {
    fn from(cfg: &config::trgt::Target) -> Self {
        use config::trgt::Target as cfg_target;
        match cfg {
            cfg_target::Dummy => Target::Dummy,
            cfg_target::Exec(e) => Target::Exec(s_exec::Exec::new(e)),
            cfg_target::GPT(g) => Target::GPT(s_gpt::GPT::new(g)),
            cfg_target::BTRFS(b) => Target::BTRFS(s_fs::btrfs::BTRFS::new(b)),
        }
    }
}
