/*
 * Copyright (c) 2021-2022, Luca Fulchir <luker@fenrirproject.org>
 *
 * This file is part of dfim.
 *
 * dfim is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * dfim is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with dfim.  If not, see <https://www.gnu.org/licenses/>.
 */

use super::NextTarget;

pub struct Exec(crate::config::trgt::exec::Exec);

impl Exec {
    pub fn new(conf: &crate::config::trgt::exec::Exec) -> Self {
        Exec(conf.clone())
    }
}

#[derive(::serde::Deserialize, Clone)]
#[serde(deny_unknown_fields)]
struct ExecNextTarget {
    id: String,
    path: ::std::path::PathBuf,
}

#[derive(::serde::Deserialize, Clone)]
#[serde(deny_unknown_fields)]
struct ExecGetCurrentResult {
    status: super::CheckStatus,
    next: Vec<ExecNextTarget>,
}

// do not add deny_unknown_fields here so that we can
// add more stuff in the future if we need it
#[derive(::serde::Deserialize, Clone)]
struct ExecDevToConfResult {
    supports_dry_run: bool,
}

#[derive(::serde::Serialize, Clone)]
#[serde(deny_unknown_fields)]
enum Action {
    DevToConfig,
    GetCurrent,
    Apply,
}

#[derive(::serde::Serialize, Clone)]
#[serde(deny_unknown_fields)]
struct CurrentDev<'a> {
    id: &'a str,
    path: &'a ::std::path::Path,
}

// used only to convert Cfg to json without the hashmaps
#[derive(::serde::Serialize, Clone)]
#[serde(deny_unknown_fields)]
struct JsonConfig<'a> {
    pub action: Action,
    pub current_device: CurrentDev<'a>,
    pub main: &'a crate::config::Main,
    pub devices: &'a Vec<crate::config::device::Device>,
    pub targets: &'a Vec<crate::config::trgt::Target>,
}

impl Exec {
    fn run(
        &mut self,
        action: Action,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<::std::process::Output, ::std::io::Error> {
        let json_cfg = ::serde_json::to_string(&JsonConfig {
            action: action,
            current_device: CurrentDev {
                id: &cfg_dev.id,
                path: dev.syspath(),
            },
            main: &cfg.main,
            devices: &cfg.devices,
            targets: &cfg.targets,
        })?;

        use std::process::Command;
        let mut cmd = match Command::new("env")
            .arg("-i")
            .arg("--")
            .arg(self.0.path.as_os_str())
            .stdin(::std::process::Stdio::piped())
            .stdout(::std::process::Stdio::piped())
            .stderr(::std::process::Stdio::piped())
            .spawn()
        {
            Ok(res) => res,
            Err(e) => {
                ::slog::error!(
                    logger,
                    "Exec: failed to run \"{:?}\"",
                    self.0.path
                );
                return Err(e);
            }
        };
        let mut cmd_stdin = cmd.stdin.take().ok_or(::std::io::Error::new(
            ::std::io::ErrorKind::BrokenPipe,
            "Exec: can't get stdin",
        ))?;
        /*
        let cmd_stdout = cmd.stdout.take().ok_or(::std::io::Error::new(
            ::std::io::ErrorKind::BrokenPipe,
            "Exec: can't get stdout",
        ))?;
        let cmd_stderr = cmd.stderr.take().ok_or(::std::io::Error::new(
            ::std::io::ErrorKind::BrokenPipe,
            "Exec: can't get stderr",
        ))?;
        */
        use std::io::Write;
        cmd_stdin.write_all(json_cfg.as_bytes())?;
        let res = match cmd.wait_with_output() {
            Ok(res) => res,
            Err(e) => {
                ::slog::error!(
                    logger,
                    "Exec \"{:?}\": error waiting for child",
                    self.0.path
                );
                return Err(e);
            }
        };
        if res.stderr.len() > 0 {
            let stderr_utf8 = match ::std::str::from_utf8(&res.stderr) {
                Ok(err_str) => err_str,
                Err(_) => {
                    ::slog::error!(
                        logger,
                        "Exec: \"{:?}\" Non utf8 stderr output",
                        self.0.path
                    );
                    ""
                }
            };
            if stderr_utf8.len() > 0 {
                for line in stderr_utf8.lines() {
                    ::slog::error!(
                        logger,
                        "Exec: \"{:?}\": {}",
                        self.0.path,
                        line
                    );
                }
            }
        }
        if !res.status.success() {
            ::slog::debug!(
                logger,
                "Exec \"{:?}\": nonzero exit status: {:?}",
                self.0.path,
                res.status.code()
            );
            return Err(::std::io::Error::new(
                ::std::io::ErrorKind::Other,
                "NonZero exit code",
            ));
        }
        Ok(res)
    }
}

// FIXME:
// This was quick to write but has multiple problems:
// * requires full buffer of the whole command run
// * separates stdout and stderr, so we don't know which message came first
// * the utf8 conversion will truncate as soon as an error is found
impl super::TargetApply for Exec {
    fn dev_to_config(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<crate::config::trgt::Target, ::std::io::Error> {
        let res = self.run(Action::DevToConfig, logger, cfg, cfg_dev, dev)?;

        let cmd_out: ExecDevToConfResult =
            match ::serde_json::from_slice(&res.stdout) {
                Err(e) => {
                    ::slog::error!(
                        logger,
                        "Exec: \"{:?}\" gave invalid or non recognised json \
                         output:\n{}",
                        self.0.path,
                        e
                    );
                    return Err(::std::io::Error::new(
                        ::std::io::ErrorKind::InvalidData,
                        "invalid json",
                    ));
                }
                Ok(cmd_out) => cmd_out,
            };
        Ok(crate::config::trgt::Target::Exec(
            crate::config::trgt::exec::Exec {
                id: "DFIM generated for ".to_owned()
                    + dev.syspath().to_str().unwrap(),
                path: self.0.path.clone(),
                supports_dry_run: cmd_out.supports_dry_run,
            },
        ))
    }
    fn get_current(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<super::GetCurrentResult, ::std::io::Error> {
        let res = self.run(Action::GetCurrent, logger, cfg, cfg_dev, dev)?;

        let cmd_out: ExecGetCurrentResult =
            match ::serde_json::from_slice(&res.stdout) {
                Err(e) => {
                    ::slog::error!(
                        logger,
                        "Exec: \"{:?}\" gave non-json output:\n{}",
                        self.0.path,
                        e
                    );
                    return Err(::std::io::Error::new(
                        ::std::io::ErrorKind::InvalidData,
                        "invalid json",
                    ));
                }
                Ok(cmd_out) => cmd_out,
            };
        let mut next =
            Vec::<super::NextTarget>::with_capacity(cmd_out.next.len());
        for n in cmd_out.next.iter() {
            let next_dev = ::udev::Device::from_syspath(&n.path)?;
            next.push(super::NextTarget {
                id: n.id.clone(),
                dev: next_dev,
            });
        }
        Ok(super::GetCurrentResult {
            status: cmd_out.status,
            conf: crate::config::trgt::Target::Exec(self.0.clone()),
            next: next,
        })
    }
    fn apply(
        &mut self,
        logger: &::slog::Logger,
        cfg: &crate::state::cfg::Cfg,
        cfg_dev: &crate::config::device::Device,
        dev: &::udev::Device,
    ) -> Result<Vec<NextTarget>, ::std::io::Error> {
        if self.0.supports_dry_run {
            let res = self.run(Action::Apply, logger, cfg, cfg_dev, dev)?;
            let out_str = match ::std::str::from_utf8(&res.stdout) {
                Ok(out_str) => out_str,
                Err(utf_err) => {
                    let idx = utf_err.valid_up_to();
                    ::std::str::from_utf8(&res.stdout[..idx]).unwrap()
                }
            };
            for line in out_str.lines() {
                ::slog::debug!(logger, "Exec: {}", line);
            }
        } else {
            ::slog::debug!(
                logger,
                "Exec: refusing to run: no dry-run support in \"{:?}\"",
                self.0.path
            );
        }
        Ok(Vec::new())
    }
}
