/*
 * Copyright (c) 2022 Luca Fulchir<luker@fenrirproject.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

mod btrfs;
mod errors;
mod swap;

use ::std::process::Stdio;
use errors::MkError;

pub enum Output {
    Quiet,
    StdoutOnly(Stdio),
    StderrOnly(Stdio),
    All { stdout: Stdio, stderr: Stdio },
}

impl Output {
    pub(crate) fn to_pair(self) -> (Stdio, Stdio) {
        match self {
            Output::Quiet => (Stdio::null(), Stdio::null()),
            Output::StdoutOnly(out) => (out, Stdio::null()),
            Output::StderrOnly(err) => (Stdio::null(), err),
            Output::All { stdout, stderr } => (stdout, stderr),
        }
    }
}

fn dev_to_node(dev: &::udev::Device) -> Result<String, MkError> {
    match dev.devnode() {
        Some(node) => match node.to_str() {
            Some(node) => Ok(node.to_owned()),
            None => {
                return Err(::std::io::Error::new(
                    ::std::io::ErrorKind::NotFound,
                    "device node path not utf8",
                )
                .into())
            }
        },
        None => {
            return Err(::std::io::Error::new(
                ::std::io::ErrorKind::NotFound,
                "device node not found",
            )
            .into())
        }
    }
}

fn run(
    mk_cmd: &str,
    mk_args: Vec<String>,
    out: ::std::process::Stdio,
    err: ::std::process::Stdio,
) -> Result<(), MkError> {
    let mut cmd = ::std::process::Command::new("env")
        .arg("-i")
        .arg("--")
        .arg(mk_cmd)
        .args(mk_args)
        .stdin(::std::process::Stdio::null())
        .stdout(out)
        .stderr(err)
        .spawn()?;
    let status = cmd.wait()?;
    match status.code() {
        Some(code) => match code {
            0 => Ok(()),
            id @ _ => Err(MkError::Exit(id)),
        },
        None => {
            use std::os::unix::prelude::ExitStatusExt;
            match status.stopped_signal() {
                Some(signal) => Err(MkError::Terminated(signal)),
                None => {
                    /* we should never get here */
                    assert!(false, "we should have found a signal");
                    Err(MkError::Terminated(::std::i32::MAX))
                }
            }
        }
    }
}
